from . import run_from_terminal

if __name__ == "__main__":
    """
    Entry point for the CLI (Command Line Interface).

    Calls the run_from_terminal function
    from the __init__.py file
    which is used for handling command line arguments.
    """
    run_from_terminal()
